// Spline-2.cc
//
// Implementation for methods of Spline<2>
//

#ifdef TEMPORARILY_OUT_OF_SERVICE

#include "splines/Spline.h"
#include "splines/Grid.h"
#include "splines/GridSearchAnswer.h"
#include "splines/DoubConv.h"
#include <vector>
#include <algorithm>
#include <cassert>
#include <sstream>
#include <iomanip>
using std::setw;

//#define TRACE_S1
//#define TRACE_S2
//#define TRACE_S3
//#define TRACE_S4
//#define TRACE_S5

namespace splines {

struct WrapperFor_2_ArgFunction {
  WrapperFor_2_ArgFunction(double ff(std::vector<double> const & v)) 
  : f(ff) 
  , v(2) { }
  double operator() (double x, double y)  const { 
    v[0] = x; 
    v[1] = y;
    return f(v); }
  mutable std::vector<double> v;
  double (*f)(std::vector<double> const & v);
};

Spline<2>::Spline ( double f(double, double), const Grid<2> & g  ) 
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
{
  captureValues ( f );
  computeDerivatives  ( );
}

Spline<2>::Spline ( double f(int x_index, int y_index), const Grid<2> & g )   
  : grid(g)
  , specifiedByTable(true)
#ifdef EMBEDDING
  , embedding()
#endif
{
  captureValues ( f );
  computeDerivatives  ( );
}

Spline<2>::Spline ( const Grid<2> & g, double *vals, double *derivs) 
  : grid(g)
#ifdef EMBEDDING
  , embedding()
#endif
{
  values = vals;
  derivatives = derivs;
  dx = derivatives;
  int n = grid.nPoints();
  dy  = dx + n;
  dxy = dy + n;  
}

Spline<2>::Spline ( double f(std::vector<double> const & v), const Grid<2> & g )
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

Spline<2>::Spline ( double f(std::vector<int> const & indicies), 
		    const Grid<2> & g )
  : grid(g)
  , specifiedByTable(false)
#ifdef EMBEDDING
  , embedding()
#endif
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

#ifdef EMBEDDING
  
Spline<2>::Spline ( double f(std::vector<double> const & v), 
		    const  Grid<2> & g,
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , specifiedByTable(true)
  , embedding( WrapperFor_2_ArgFunction(outerF) )  
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

 				   
Spline<2>::Spline ( double f(std::vector<int> const & indicies), 
		    const  Grid<2> & g, 
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , specifiedByTable(true)
 , embedding( WrapperFor_2_ArgFunction(outerF) )  
{  
  captureValues ( f );
  computeDerivatives  ( ); 
}

Spline<2>::Spline ( const  Grid<2> & g, 
		    double *vals, 
		    double *derivs,
  	   	    double outerF(std::vector<double> const & v ) )
  : grid(g)
  , embedding( WrapperFor_2_ArgFunction(outerF) )  
{ 
  values = vals;
  derivatives = derivs; 
  dx = derivatives;
  int n = grid.nPoints();
  dy  = dx + n;
  dxy = dy + n;  
}
#endif

#ifdef TODO
void
Spline<1>::setTables (std::istream & is) {
  is >> grid;
  unsigned int N = grid.nPoints();
  std::vector<unsigned long> v(2);
  for (unsigned int i=0; i<N; ++i) {
    is >> v[0] >> v[1];
    values_vector.push_back ( DoubConv::longs2double(v) );
  }
  for (unsigned int j=0; j<N; ++j) {
    is >> v[0] >> v[1];
    derivatives_vector.push_back ( DoubConv::longs2double(v) );
  }
  values      = & values_vector[0];
  derivatives = & derivatives_vector[0];
}
#endif

Spline<2>::Spline (const Spline<2> & s) : grid(s.grid), 
#ifdef EMBEDDING
					  embedding(s.embedding),
#endif
			 		  values(s.values), 
					  derivatives(s.derivatives),
					  dx(s.dx),
					  dy(s.dy),
					  dxy(s.dxy) { }

void 
Spline<2>::captureValues ( double f(double, double) ) {
  std::vector<int> shape = grid.shape();
  nx = shape[0];
  ny = shape[1];
  values_vector.clear(); values_vector.reserve( nx*ny );
  std::vector<double> gx = grid.points(0);
  std::vector<double> gy = grid.points(1);
  double x;
  double y;
  for (unsigned int ix = 0; ix < nx; ++ix) {
   x = gx[ix];
   for (unsigned int iy = 0; iy < ny; ++iy) {
     y = gy[iy];
     values_vector.push_back ( f(x,y) );
   }
  }
  values = &values_vector[0];
  specifiedByTable = false;
  boundaryDerivatives(f);
}

void 
Spline<2>::captureValues(double f(int x_index, int y_index) ) {
  std::vector<int> shape = grid.shape();
  nx = shape[0];
  ny = shape[1];
  values_vector.clear(); values_vector.reserve( nx*ny );
  for (unsigned int ix = 0; ix < nx; ++ix) {
   for (unsigned int iy = 0; iy < ny; ++iy) {
     values_vector.push_back ( f(ix,iy) );
   }
  }
  values = &values_vector[0];
  specifiedByTable = true;
}

void 
Spline<2>::captureValues ( double f(std::vector<double> const &) ){
  std::vector<int> shape = grid.shape();
  int nx = shape[0];
  int ny = shape[1];
  values_vector.clear(); values_vector.reserve( nx*ny );
  std::vector<double> gx = grid.points(0);
  std::vector<double> gy = grid.points(1);
  std::vector<double> v(2);
  for (unsigned int ix = 0; ix < nx; ++ix) {
   v[0] = gx[ix];
   for (unsigned int iy = 0; iy < ny; ++iy) {
     v[1] = gy[iy];
     values_vector.push_back ( f(v) );
   }
  }
  values = &values_vector[0];
  specifiedByTable = false;
  boundaryDerivatives( WrapperFor_2_ArgFunction (f) );
}

void 
Spline<2>::captureValues( double f(std::vector<int> const &) ){
  std::vector<int> shape = grid.shape();
  int nx = shape[0];
  int ny = shape[1];
  values_vector.clear(); values_vector.reserve( nx*ny );
  std::vector<int> v(2);
  for (unsigned int ix = 0; ix < nx; ++ix) {
   v[0] = ix;
   for (unsigned int iy = 0; iy < ny; ++iy) {
     v[1] = iy;
     values_vector.push_back ( f(v) );
   }
  }
  values = &values_vector[0];
  specifiedByTable = true;
}


void 
Spline<2>::computeDerivatives ( ) {
  std::vector<double> gx = grid.points(0);
  std::vector<double> gy = grid.points(1);
  int Nx = gx.size();
  int Ny = gy.size();
  int n = grid.nPoints();
  derivatives_vector.resize(3*n);  // 3 for dx, dy, dxy
  dx = &derivatives_vector[0];
  dy = dx + n;
  dxy = dy + n; 
  // Compute the x derivatives
  std::vector<double> xvals(Nx);
  for (int iy = 0; iy < Ny; ++iy) {
    for (int p = 0; p < Nx; ++p) {
      xvals[p] = values[iy+Ny*p];
    }
    if (specifiedByTable) {
      computeLineOfDerivatives (gx, xvals, 
      				dx+iy,  // Where to place first one
				Ny);	// stride
    } else {
      computeLineOfDerivatives (gx, xvals, 
      				dx+iy,  // Where to place first one
				Ny,	// stride
				x_slope0[iy], 
				x_slopeN[iy]);
    }
  }
  // Compute the y derivatives
  std::vector<double> yvals(Ny);
  for (int ix = 0; ix < Nx; ++ix) {
    for (int p = 0; p < Ny; ++p) {
      yvals[p] = values[ix*Nx+p];
    }
    if (specifiedByTable) {
      computeLineOfDerivatives (gy, yvals, 
      				dy+ix*Nx,  // Where to place first one
				1);	   // stride
    } else {
      computeLineOfDerivatives (gy, yvals, 
      				dy+ix*Nx,  // Where to place first one
				1,	   // stride
				y_slope0[ix], 
				y_slopeN[ix]);
    }
  }
  // Compute the mixed second derivatives along the y-edges
  // by computing the x derivative of the y derivative
  for (int iy = 0; iy < Ny; iy+=Ny-1) {
    for (int p = 0; p < Nx; ++p) {
      xvals[p] = dy[iy+Ny*p];
    }
    if (specifiedByTable) {
      computeLineOfDerivatives (gx, xvals, 
      				dxy+iy,  // Where to place first one
				Ny);	 // stride
    } else {
      double slope0 = ( (iy==0) ? dxy00 : dxy0N );
      double slopeN = ( (iy==0) ? dxyN0 : dxyNN );
      computeLineOfDerivatives (gx, xvals, 
      				dxy+iy,  // Where to place first one
				Ny,	 // stride
				slope0, 
				slopeN);
    }
  }
  // Compute the mixed second derivatives everywhere
  // by computing the y derivative of the x derivative,
  // using the second derivatives along the y-edges 
  // as boundary values of slopes
  for (int ix = 0; ix < Nx; ++ix) {
    for (int p = 0; p < Ny; ++p) {
      yvals[p] = dx[ix*Nx+p];
    }
    if (specifiedByTable) {
      computeLineOfDerivatives (gy, yvals, 
      				dxy+ix*Nx,  // Where to place first one
				1);	    // stride
    } else {
      double slope0 = dxy[ix*Ny];
      double slopeN = dxy[(ix+1)*Ny-1];
      computeLineOfDerivatives (gy, yvals, 
      				dy+ix*Nx,  // Where to place first one
				1,	   // stride
				slope0, 
				slopeN);
    }
  }
} // computeDerivatives ( ) 

void
Spline<2>::swap (Spline<2> & s) {
  grid.swap(s.grid);
  values_vector.swap(s.values_vector);
  std::swap(values, s.values);
  if ( !  values_vector.empty() )   values = &  values_vector[0];
  if ( !s.values_vector.empty() ) s.values = &s.values_vector[0];
  derivatives_vector.swap(s.derivatives_vector);
  std::swap(derivatives, s.derivatives);
  std::swap(dx , s.dx );
  std::swap(dy , s.dy );
  std::swap(dxy, s.dxy);
  if ( !  derivatives_vector.empty() )   {
    int n = grid.nPoints();
    derivatives = &  derivatives_vector[0];
    dx = derivatives;
    dy = dx + n;
    dxy = dy + n;
  }
  if ( !s.derivatives_vector.empty() )   {
    int n = s.grid.nPoints();
    s.derivatives = &s.derivatives_vector[0];
    s.dx = s.derivatives;
    s.dy = s.dx + n;
    s.dxy = s.dy + n;
  }
}

Spline<2> &
Spline<2>::operator= (const Spline<2> & s) {
  Spline<2> temp(s);
  swap(temp);
  return *this;
}


void
Spline<2>::computeLineOfDerivatives( std::vector<double> const & x,
				     std::vector<double> const & values,
  				     double * dest,
				     int stride,
				     double slope0,
				     double slopeN ) {
  int N = x.size();
  std::vector<double> y2 (N);
  double s0 = 3 * ( (values[1]-values[0])/(x[1]-x[0]) - slope0 ) 
							/ (x[1]-x[0]);
  double sN = 3 * ( slopeN - (values[N-1]-values[N-2])/(x[N-1]-x[N-2]) ) 
							/ (x[1]-x[0]);
  y2 = secondDerivatives (x, &values[0], .5, s0, .5, sN ); 
  
  // apply 3.3.5, with A =1 and B = 0 for all points except the last:
  
  for (int i=0; i<N-1; ++i) {
    *dest = (   (values[i+1]-values[i])/(x[i+1]-x[i])
			           - (x[i+1]-x[i]) * (2*y2[i] + y2[i+1])/6.0  );
    dest += stride;
  }
  // For the last point, use A = 0 and B = 1 in the last interval:
  *dest =  (   (values[N-1]-values[N-2])/(x[N-1]-x[N-2])
			         + (x[N-1]-x[N-2]) * (y2[N-2] + 2*y2[N-1])/6.0);
}				     

void
Spline<2>::computeLineOfDerivatives( std::vector<double> const & x,
				     std::vector<double> const & values,
  				     double * dest,
				     int stride ) {
  int N = x.size();
  std::vector<double> y2 (N);
  y2 = secondDerivatives (x, &values[0], 0, 0, 0, 0); 
  
  // apply 3.3.5, with A =1 and B = 0 for all points except the last:
  
  for (int i=0; i<N-1; ++i) {
    *dest = (   (values[i+1]-values[i])/(x[i+1]-x[i])
			           - (x[i+1]-x[i]) * (2*y2[i] + y2[i+1])/6.0  );
    dest += stride;
  }
  // For the last point, use A = 0 and B = 1 in the last interval:
  *dest =  (   (values[N-1]-values[N-2])/(x[N-1]-x[N-2])
			         + (x[N-1]-x[N-2]) * (y2[N-2] + 2*y2[N-1])/6.0);
}				     


std::vector<double> 
Spline<2>::
secondDerivatives (const std::vector<double> & x, 
		   const double* y,
		   double a0, double s0,
		   double bN, double sN) {

  int N = x.size();
  std::vector<double> y2 (N);

  // We will be solving M y2 = s, where M is a tri-diagonal matrix.
  // $$$ See Mathematical Details 1.2 $$$ 

  std::vector<double> m (N);  // Diagonal elements of M 
  std::vector<double> a (N);  // Elements of M above the diagonal (0 - N-2)
  std::vector<double> b (N);  // Elements of M below the diagonal (1 - N-1)
  std::vector<double> s (N);  
  // Optimization opportunity:
  //
  // This can be done with only two vector of temp storage space, s and a.
  // For code clarity, we use these four instead.

  // Step 1: Form M and s 

  if ( a0 != 0 ) {	// derivative Boundary condition at 0
    assert (a0 == .5);
  } else {
    s0 = 0;
  }

  if ( bN != 0 ) {	// derivative Boundary condition at N-1
    assert (bN == .5);
  } else {
    sN = 0;
  }

  m[0] = 1;
  a[0] = a0;
  s[0] = s0;

  int i;
  for ( i = 1; i < N-1; ++i ) {
    b[i] = ( x[i]   - x[i-1] ) /6.0;
    m[i] = ( x[i+1] - x[i-1] ) /3.0;
    a[i] = ( x[i+1] - x[i]   ) /6.0;
    s[i] = (y[i+1] - y[i])/(x[i+1] - x[i]) - (y[i] - y[i-1])/(x[i] - x[i-1]); 
  }

  b[N-1] = bN;
  m[N-1] = 1;
  s[N-1] = sN;


  // Step 2: Eliminate the lower diagonal elements; m[i] becomes implied 1's

  double multiplier;
  for ( i = 0; i < N-1; ++i ) {
    multiplier = 1.0/m[i];
    a[i] *= multiplier;
    s[i] *= multiplier;
    m[i+1] -= b[i+1] * a[i];
    s[i+1] -= b[i+1] * s[i];
  }
  multiplier = 1.0/m[i];
  s[N-1] /= m[N-1];

  // Step 3: Backsolve (eliminate the upper diagonal a[i])

  for ( i = N-1; i > 0; --i ) {
    s[i-1] -= a[i-1] * s[i];
  }

  return s;

}


#ifdef TODO
double 
Spline<1>::operator() ( double xval ) const {

  // $$$ See Mathematical Details 1.1 $$$ 

  // Find the interval that xval lies within.  
  // $$$ For out-of-range cases, see Mathematical Details 1.4 $$$

  GridSearchAnswer a;
  bool interior = grid.search(xval, a);
				#ifdef TRACE_S1
				static int ntraces=0;
				if (ntraces < 10) { 
				std::cout << "Spline (" << xval << "): \n" 
				<< "x = " << a.x << " h = " << a.h 
				<< " a.index = " << a.index 
				<< "\nvalues " 	<<  values[a.index] 
				<<  ", " << values[a.index+1] 
				<< " grads  " <<  derivatives[a.index] << ", " 
				<< derivatives[a.index+1] << "\n";
				}
				#endif

  // Compute the f0, f1, g0 and g1 functions. 
  // $$$ See Mathematical Details 1.5 $$$

//  if ( (!interior) && embedding ) { 
//    return embedding(xval); 
//  }
  
#ifdef EMBEDDING
   if (!interior) {
    if (embedding) { 
      return embedding(xval);
    } 
  }
#endif
  
  double x = a.x;
  double oneMinusX = 1.0 - x;

  double spline;

  double x2 = x*x;
  double oneMinusX2 = oneMinusX * oneMinusX;
  double h = a.h;

  double f0 = (2. * x + 1.) * oneMinusX2;
  double f1 = 1.0 - f0;				// same as (3. - 2. * x) * x2;
  double g0 =   h * x  * oneMinusX2;
  double g1 =  -h * x2 * oneMinusX; 
				#ifdef TRACE_S1
				if (ntraces < 10) { ntraces++;
				std::cout << " f0 = " << f0 
		  			  << " f1 = " << f1
		  			  << " g0 = " << g0
		  			  << " g1 = " << g1 << "\n";
				}
				#endif 

  spline =   f0 * values[a.index]       + f1 * values[a.index+1]
	   + g0 * derivatives[a.index]  + g1 * derivatives[a.index+1] ;

  return spline;
 
} /* operator() */
#endif

#ifndef TODO_THIS_IS_TEMPORARY
double 
Spline<2>::operator() ( double xval, double yval ) const {
  return 0;
}
#endif

double 
Spline<2>::operator() ( const std::vector<double> & v ) const {
  if ( v.size() != 2 ) throw  Spline_argument_exception(2,v.size());
  return operator() ( v[0], v[1] );
}

void 
Spline<2>::save (std::vector<double> & values_output, 
		 std::vector<double> & derivatives_output) const {
  values_output      = values_vector;
  derivatives_output = derivatives_vector;
}		 

bool
Spline<2>::operator==( Spline<2> const & rhs ) const {
  if (grid != rhs.grid) return false;
  unsigned int N = grid.nPoints();
  for (unsigned int i = 0;  i < N; ++i) {
    if (values[i]      != rhs.values[i]     ) return false;
    if (derivatives[i] != rhs.derivatives[i]) return false;
  }
  return true; 
}

std::ostream &
Spline<2>::put (std::ostream & os) const {
  os << grid;
  os << "\n";
  unsigned int N = grid.nPoints();
  std::vector<unsigned long> v(2);
  for (unsigned int i=0; i<N; ++i) {
    v = DoubConv::dto2longs(values[i]);
    os << v[0] << " " << v[1] << "\n";
  }
  os << "\n";
  for (unsigned int j=0; j<N; ++j) {
    v = DoubConv::dto2longs(derivatives[j]);
    os << v[0] << " " << v[1] << "\n";
  }
  return os;  
}


} // end of namespace splines

#endif // TEMPORARILY_OUT_OF_SERVICE
